import datetime
import inspect
from collections.abc import Callable
from functools import partial, reduce
from sys import argv, exit

from colored import Back, Fore, Style

from .tcr_dict import clean_dunder_dict
from .tcr_extract_error import extract_error, extract_traceback
from .tcr_getch import getch
from .tcr_inspect import eval_fback, get_file_colon_lineno
from .tcr_iterable import cut_at
from .tcr_print import FMTC, fmt_iterable
from .tcr_terminal import terminal
from .tcr_types import QuotelessString


class CC:
	_ = Style.reset
	LOG = Fore.light_green + Style.bold  # 10
	WARN = Fore.YELLOW + Style.bold
	ERROR = Fore.RED + Style.bold
	DEBUG = Fore.purple_1b + Style.bold  # 129
	CRITICAL = Back.RED + Style.bold
	BANG = Fore.WHITE + Back.RED + Style.bold


class Console:
	"""### Provides logging capabilities.

	`console(...)` == `console.debug(...)`
	"""

	header_exprs: list[tuple[str, bool]]
	header_joiner: str
	header_end: str
	_last_diff: str | None = None
	include_callsite: bool = None

	def __init__(self, *, header_exprs: list[tuple[str, bool]] = None, header_joiner: str = ".", header_end: str = ": ") -> None:
		self.header_exprs = header_exprs or []
		self.header_joiner = header_joiner
		self.header_end = header_end

	@staticmethod
	def _get_timestamp():
		return str(datetime.datetime.now())[:-3].replace(".", ",")

	def _get_callsite_text_if_enabled(self, backtrack_frames) -> str:
		if self.include_callsite or (self.include_callsite is None and "--tcr-c-callsite" in argv):
			return get_file_colon_lineno(backtrack_frames=backtrack_frames)
		return ""

	def evaluate_header(self, fback_count: int = 5) -> str:
		result_list = [(expr if is_literal else eval_fback(expr, fback_count=fback_count)) for expr, is_literal in self.header_exprs]
		result_str = self.header_joiner.join(str(x) for x in result_list if x is not None)
		if result_str:
			result_str += self.header_end

		return result_str

	def _generate_out_and_print(
		self, *values, sep="\n", end="", withprefix=True, syntax_highlighting: bool = True, color: str, letter: str, _this_iteration_header: str | None = None, **kwargs
	) -> None:
		if _this_iteration_header is None:
			_this_iteration_header = self.evaluate_header()

		if not values:
			values = ("",)

		if len(values) > 1:
			for i, v in enumerate(values):
				if i == 0:
					char = "V"
				elif i == len(values) - 1:
					char = "Λ"
				else:
					char = "│"

				print(char, end=" ")

				self._generate_out_and_print(
					v, sep=sep, end=end, withprefix=withprefix, syntax_highlighting=syntax_highlighting, color=color, letter=letter, _this_iteration_header=_this_iteration_header, **kwargs
				)
			return

		values = [(f'{_this_iteration_header or ""}{x}' if isinstance(x, str) else fmt_iterable(x, syntax_highlighting=syntax_highlighting, **kwargs)) for x in values]

		out = sep.join(values)
		out = reduce(lambda x, y: str(x) + sep + str(y), [*values, ""]) + end

		if withprefix:
			out = f'{letter} {(self._get_callsite_text_if_enabled(4) + " ").lstrip()}{self._get_timestamp()} ' + out

		out = f'{color if syntax_highlighting else ""}{out}{CC._ if syntax_highlighting else ""}'

		print(out)

	def log(self, *values, sep=" ", end="", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.LOG, letter="I")

	def warn(self, *values, sep=" ", end="", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.WARN, letter="W")

	def error(self, *values, sep=" ", end="", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.ERROR, letter="E")

	def critical(self, *values, sep=" ", end="", withprefix=True) -> None | str:
		self._generate_out_and_print(*values, sep=sep, end=end, withprefix=withprefix, color=CC.CRITICAL, letter="C")

	def debug(
		self,
		value: object = ...,
		/,
		*values: object,
		withprefix: bool = True,
		passthrough: bool = False,
		printhook: Callable[[str], None] = print,
		syntax_highlighting=True,
		margin: str = "",
		padding: str = " ",
		quoteless: bool = True,
		diff: bool = False,
		fmt_iterable: Callable[..., str] = fmt_iterable,
		**kwargs,
	) -> None | object:
		all_values = (value, *values)

		eval_header = self.evaluate_header(6)

		if eval_header:
			padding += eval_header

		if len(all_values) >= 2 and all_values[0].__class__ == str and all_values[0]:  # noqa: E721
			first_string: str = all_values[0]
			after_first_string = ""

			for symbol in ("=", "->", "=>"):
				if first_string.rstrip().endswith(symbol):
					first_string = first_string.rstrip().removesuffix(symbol)
					if first_string and first_string[-1] == " ":
						after_first_string = f"{symbol} "
					else:
						after_first_string = symbol

			padding += fmt_iterable(QuotelessString(first_string), syntax_highlighting=syntax_highlighting, **kwargs)
			padding += after_first_string

			all_values = all_values[1:]

		out = fmt_iterable(*[(x if ((not quoteless) or (x.__class__ != str) or x == "") else QuotelessString(x)) for x in all_values], syntax_highlighting=syntax_highlighting, **kwargs)  # noqa: E721

		if padding == " " and not withprefix:
			padding = ""

		prefix = ""
		if withprefix:
			prefix = f'D {(self._get_callsite_text_if_enabled(3) + " ").lstrip()}{self._get_timestamp()}'

		c_debug = ""
		c_reset = ""
		c_bang = ""
		if syntax_highlighting:
			c_debug = CC.DEBUG
			c_reset = CC._
			c_bang = CC.BANG

		def compose(x) -> str:
			return f"{c_debug}{margin}{prefix}{padding}{x}{c_reset}"

		if diff and not self._last_diff:
			printhook(compose("\n" + "\n".join(["  " + x for x in out.split("\n")])))
		else:
			if not diff:
				printhook(compose(out))
			else:
				out_lines = out.split("\n")
				last_lines = self._last_diff.split("\n")

				while len(out_lines) < len(last_lines):
					last_lines.pop(-1)

				while len(last_lines) < len(out_lines):
					last_lines.append("")

				diff_out = "\n".join([("  " if line == line_old else f"{c_bang}*{c_reset} ") + line for line, line_old in zip(out_lines, last_lines, strict=True)])

				printhook(compose("\n" + diff_out))

		self._last_diff = out
		return None if not passthrough else value

	def hr(self, newlines_on_both_sides: bool = True, **kwargs):
		newline = "\n" if newlines_on_both_sides else ""
		self.debug(QuotelessString(newline + ("=" * terminal.width) + newline), withprefix=False, **kwargs)

	def with_expr_header(self, header_expr: str, literal: bool = False, joiner: str | None = None, end: str | None = None) -> "Console":
		if not isinstance(header_expr, str):
			raise TypeError("header_expr must be a string")

		if joiner is not None:
			if not isinstance(joiner, str):
				raise TypeError("overwrite_header_joiner must be a string")

		if end is not None:
			if not isinstance(end, str):
				raise TypeError("overwrite_header_end must be a string")

		return Console(
			header_exprs=self.header_exprs + [(header_expr, literal)],
			header_joiner=joiner if joiner is not None else self.header_joiner,
			header_end=end if end is not None else self.header_end,
		)

	__call__ = debug

	def __or__(self, other):
		return self.debug(other, passthrough=True)

	def __ror__(self, other):
		return self.debug(other, passthrough=True)


console = Console()
"""### Provides logging capabilities.

  `console(...)` == `console.debug(...)`
"""


def _clean_built_in_methods(d: dict) -> dict:
	return {k: v for k, v in d.items() if not isinstance(v, type(abs))}


def start_eval_session(f_backs: int = 2) -> None:
	"""Starts an interactive shell. For debugging purposes."""
	try:
		current_frame = inspect.currentframe()
		for _ in range(f_backs):
			current_frame = current_frame.f_back
		desired_locals = current_frame.f_locals
		desired_globals = current_frame.f_globals

		a = " "
		while a:
			a = input("\n\r>>> ")
			if a:
				try:
					if a.strip().lower() == "locals":
						console(_clean_built_in_methods(clean_dunder_dict(desired_locals)))
					elif a.strip().lower() == "globals":
						console(_clean_built_in_methods(clean_dunder_dict(desired_globals)))
					elif a.strip().lower() == "code":
						lines, start_line = inspect.findsource(current_frame)
						lineno = current_frame.f_lineno

						AROUND = 2
						CUTOFF = 70

						lines = lines[max(0, lineno - (AROUND + 1)) : lineno + AROUND]
						while all(x.startswith(" ") for x in lines):
							lines = [x[1:] for x in lines]
						lines = [f'{FMTC.NUMBER}{lineno - AROUND + i}{Fore.yellow + Style.bold} {">" if i == AROUND else "|"}{FMTC._} {cut_at(x, CUTOFF)}' for i, x in enumerate(lines)]
						lines = "".join(lines)

						print(lines)
					else:
						console(eval(a, desired_locals, desired_globals))
				except Exception as e:
					console.error(f"\n\n{extract_traceback(e)}\n{extract_error(e)}")
	except (KeyboardInterrupt, EOFError):
		print("\r" + (" " * terminal.width) + "\r", end="")
		exit()


def breakpoint(*vals, printhook=console, clear=True, ctrlc: Callable[[], None] = start_eval_session) -> None:
	"""Stop the program execution until a key is pressed. Optionally pass in things to print."""
	vals and printhook(*vals)
	print(
		(a := f"{Fore.WHITE + Style.BOLD} >>> {Back.RED}BREAKPOINT{Style.RESET} {Fore.WHITE + Style.BOLD}<<<{Style.RESET}"),
		end="\r",
	)

	if getch() == b"\x03":
		if clear:
			print(len(a) * " ", end="")
		ctrlc()
		return

	if clear:
		print(len(a) * " ", end="\r")
	else:
		print()



import pathlib as p
import sys
import threading as th
import traceback
import warnings as ws
from collections.abc import Callable

from colored import Back, Fore, Style

from .tcr_classfuncs import get_classname
from .tcr_console import console
from .tcr_iterable import cut_at
from .tcr_print import FMTC

C_WHITE = Fore.white + Style.bold
C_NUMBER = FMTC.NUMBER
C_WARNING = Fore.yellow + Style.bold
C_ERROR = Fore.red + Style.bold
C_EXCEPTION = FMTC.INTERNAL_EXCEPTION


def default_err_or_warning_printhook(s: str, warning: Warning | Exception, filename: str, lineno: int, *, tb=None) -> None:
	"""Default printhook for tcr.WarningCatcher."""
	path = p.Path(filename)

	AROUND: int = 2
	"""How many lines around the line of the warning to show."""
	CUTOFF: int = 70
	"""How many (max) characters can be shown per line until they ge..."""

	is_warning = isinstance(warning, Warning)

	color = C_WARNING if is_warning else C_ERROR

	if not path.exists():
		lines = ""
	else:
		lines = path.read_text().split("\n")
		lines = lines[max(0, lineno - (AROUND + 1)) : lineno + AROUND]
		while all(x.startswith(" ") for x in lines):
			lines = [x[1:] for x in lines]
		lines = [f'{C_NUMBER}{lineno - AROUND + i}{color} {">" if i == AROUND else "|"}{FMTC._} {cut_at(x, CUTOFF)}' for i, x in enumerate(lines)]
		lines = "\n".join(lines)
		lines = f"\n{lines}"

	if tb:
		traceback.print_tb(tb)

	func = console.warn if is_warning else console.error
	func(f'{color}{filename}{C_WHITE}:{C_NUMBER}{lineno}{C_WHITE}: {C_EXCEPTION}{get_classname(warning)}{C_WHITE}: {color}{str(warning).replace("<locals>.", "")}{lines}')


def _errorcatcher_excepthook(type, value, traceback):
	frame = traceback.tb_frame
	filename = frame.f_code.co_filename
	lineno = frame.f_lineno
	default_err_or_warning_printhook(str(value), value, filename, lineno, tb=traceback)


class ErrorCatcher:
	def __init__(self, **kwargs) -> None:
		self.install(**kwargs)

	@staticmethod
	def install():
		sys.excepthook = _errorcatcher_excepthook


class WarningCatcher:
	"""Catches any warnings (for example "Coroutine X was not awaited") and forwards them as text to the selected (or default: tcr.console.warn) printhook.

	Usage:
	```py
	tcr.WarningCatcher()
	# OR
	tcr.WarningCatcher.install() # Tgus
	```
	"""

	@staticmethod
	def install(**kwargs):
		"""Alias for tcr.WarningCatcher(**kwargs). Use it for better typehints."""
		_WarningCatcher.install(**kwargs)

	def __init__(
		self,
		printhook: Callable[[str, Warning, str, int], None] = default_err_or_warning_printhook,
	) -> None:
		"""Catches any warnings (for example "Coroutine X was not awaited") and forwards them as text to the selected (or default: tcr.console.warn) printhook.

		Usage:
		```py
		tcr.WarningCatcher()
		```

		Args:
		        - printhook: Callable[[str, Warning, str, int], None], Defines a printer function, for more info see this module's default_showwarning_printhook function and https://docs.python.org/3/library/warnings.html#warnings.showwarning.
		"""
		_WarningCatcher.install(
			printhook=printhook,
		)


class _WarningCatcher(th.Thread):
	_initiated = False
	_installed = False

	def __init__(self, *, printhook):
		if _WarningCatcher._initiated:
			raise RuntimeError("WarningCatcher can only be initiated once.")
		super().__init__()
		_WarningCatcher._initiated = True
		self.__printhook = printhook
		self.daemon = True
		self._stop_event = th.Event()

	def run(self):
		def showwarning(message, category, filename, lineno, file=None, line=None):
			self.__printhook(str(message), message, filename, lineno)

		with ws.catch_warnings():
			ws.showwarning = showwarning
			while not self._stop_event.is_set():
				self._stop_event.wait(0.1)  # Check every 0.1 second for the stop signal

	def stop(self):
		self._stop_event.set()

	@classmethod
	def install(cls, **kwargs):
		if _WarningCatcher._installed:
			raise RuntimeError("WarningCatcher can only be installed once.")
		_WarningCatcher._installed = True

		catcher = cls(**kwargs)
		catcher.start()
		return catcher



import contextlib
import inspect
import traceback
from collections.abc import Callable
from types import ModuleType

from .tcr_compare import able
from .tcr_iterable import getattr_queue


def extract_error(e: BaseException, pattern="%s: %s", *, raw=False) -> tuple[str, str] | str:
	if able(issubclass, e, BaseException) and issubclass(e, BaseException):
		tup = (getattr_queue(e, "__name__", "__class__.__name__", default="!UnknownExceptionGroupClass"), "<class>")
		return tup if raw else pattern % tup

	if able(issubclass, e, Exception) and issubclass(e, Exception):
		tup = (getattr_queue(e, "__name__", "__class__.__name__", default="!UnknownExceptionClass"), "<class>")
		return tup if raw else pattern % tup

	if callable(e):
		e = e()

	one = getattr_queue(e, "__name__", "__class__.__name__", default="!UnknownException")
	two = str(e)
	return (one, two) if raw else pattern % (one, two)


def extract_traceback(e: BaseException) -> str:
	traceback_details = traceback.format_tb(e.__traceback__)
	return "".join(traceback_details)


def print_exception_with_traceback(e: BaseException, *, printhook: Callable[[str], None] = print) -> None:
	printhook("Traceback (most recent call last):")
	printhook(extract_traceback(e))
	printhook(extract_error(e))


def module_error_map(module: ModuleType) -> dict[str, BaseException]:
	"""Get all exception classes in a module."""
	return {name: obj for name, obj in vars(module).items() if inspect.isclass(obj) and issubclass(obj, BaseException) and not issubclass(obj, Warning)}


def modules_error_map(*modules: ModuleType) -> dict[str, BaseException]:
	"""Get all exception classes in a list of modules."""
	return {f"{module.__name__}.{name}".removeprefix("builtins."): exc for module in modules for name, exc in module_error_map(module).items()}



from .. import util as __


@__.STRINGIFY
@__.FLATTEN_IF_POSSIBLE
def add(_, *args: str, **ctxs) -> int | float:
	"""### Add numbers (supports any number of args).

	Example:
	```txt
	{add|1|2|3}
	```

	Result:
	```txt
	6
	```
	"""
	args = [__.number(x, flatten_to_int_when_possible=True) for x in args]
	return sum(args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 0)
@__.FLATTEN_IF_POSSIBLE
def subtract(_, *args: str, **ctxs) -> int | float:
	"""### Subtract numbers (supports any number of args).

	Example:
	```txt
	{add|1|2|3}{#|evaluates to putting a '-' between each number: 1 - 2 - 3}
	```

	Result:
	```txt
	-4
	```
	"""
	args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

	if len(args) < 2:
		return args[0]

	return __.functools.reduce(lambda x, y: x - y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 1)
@__.FLATTEN_IF_POSSIBLE
def multiply(_, *args: str, **ctxs) -> int | float:
	"""### Multiply numbers (supports any number of args).

	Example:
	```txt
	{multiply|2|2|3}
	```

	Result:
	```txt
	12
	```
	"""
	args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

	if len(args) < 2:
		return args[0]

	return __.functools.reduce(lambda x, y: x * y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 1)
@__.FLATTEN_IF_POSSIBLE
def divide(_, *args: str, **ctxs) -> int | float:
	"""### Divide numbers (supports any number of args).

	Example:
	```txt
	{divide|1|2}{#|in case of >2 arguments evaluates to putting a '/' between each number: 1 / 2 / 3}
	```

	Result:
	```txt
	0.5
	```
	"""
	args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

	if len(args) < 2:
		return args[0]

	return __.functools.reduce(lambda x, y: x / y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, 1)
@__.FLATTEN_IF_POSSIBLE
def floordivide(_, *args: str, **ctxs) -> int | float:
	"""### Floor divide numbers (supports any number of args).

	Example:
	```txt
	{floordivide|34|3}{#|in case of >2 arguments evaluates to putting a '//' between each number: 1 // 2 // 3}
	```

	Result:
	```txt
	11
	```
	"""
	args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

	if len(args) < 2:
		return int(args[0])

	return __.functools.reduce(lambda x, y: x // y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
@__.FLATTEN_IF_POSSIBLE
def power(_, base: str, exponent: str = "2", *args: str, **ctxs) -> int | float:
	args = [__.number(x, flatten_to_int_when_possible=True) for x in (base, exponent, *args)]

	return __.functools.reduce(lambda x, y: x**y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
@__.FLATTEN_IF_POSSIBLE
def modulo(_, *args: str, **ctxs) -> int | float:
	"""### Modulo numbers (supports any number of args).

	Example:
	```txt
	{modulo|34|3}{#|in case of >2 arguments evaluates to putting a '%' between each number: 1 % 2 % 3}
	```

	Result:
	```txt
	1
	```
	"""
	args = [__.number(x, flatten_to_int_when_possible=True) for x in args]

	if len(args) < 2:
		return args[0]

	return __.functools.reduce(lambda x, y: x % y, args)


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
def round_(_, n: str, *args: str, **ctxs) -> int | float:
	"""### Round numbers using python's round function. (This DISCARDS any extra arguments!).

	This does not just round halves up, see: https://en.wikipedia.org/wiki/Rounding#Rounding_half_to_even

	Example:
	```txt
	{round|1.2}
	{round|1.5}
	{round|2.5}
	{round|3.5}
	{round|4.5}
	```

	Result:
	```txt
	1
	2
	2
	4
	4
	```
	"""
	return round(__.number(n))


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
def floor(_, n: str, *args: str, **ctxs) -> int | float:
	"""### Round down. (This DISCARDS any extra arguments!).

	Example:
	```txt
	{floor|6.1}
	{floor|6.9}
	```

	Result:
	```txt
	6
	6
	```
	"""
	return __.math.floor(__.number(n))


@__.STRINGIFY
@__.REQUIRE_POSITIONAL(1, None)
def ceil(_, n: str, *args: str, **ctxs) -> int | float:
	"""### Round up. (This DISCARDS any extra arguments!).

	Example:
	```txt
	{ceil|9.1}
	{ceil|9.9}
	```

	Result:
	```txt
	10
	10
	```
	"""
	return __.math.ceil(__.number(n))



from types import UnionType
from typing import Any, Self, TypeVar, Union

from .tcr_class import Singleton


def instance(func):
	return func()


T = TypeVar("T")


class _NullLike:
	"""Base class for null-like objects (Null, Undefined).

	Can be bitwise-ored with any other value to yield that other value. (If both values in bitwise or are null-like, return the directionally right one)
	"""

	def __str__(self):
		return self.__repr__()

	def __repr__(self):
		return self.__class__.__name__

	def __bool__(self):
		return False

	def __or__(self, __value: T) -> T:
		return __value

	def __ror__(self, __value: T) -> T:
		return __value


@instance
class Null(Singleton, _NullLike):
	"""Value is explicitly declared as absent."""


@instance
class Undefined(Singleton, _NullLike):
	"""Value is missing."""


class UniqueDefault:
	"""Used as a default parameter for the parameter named default (yeah i know that's uh... let's just see the example).

	Example:
	```py
	from tcrutils import UniqueDefault as RaiseError

	def get_something(..., *, default: Any | RaiseError = RaiseError):
	... # Function implementation

	# Nothing was found, falling to raising error or returning default value if such was specified.
	if default is not RaiseError:
	  return default
	raise KeyError("Unable to find something.")
	"""



from __future__ import annotations as __future_annotations

from ...discord.tcrd_snowflake import is_snowflake as __is_snowflake
from ...discord.tcrd_string import IFYs
from ...src.tcr_console import console as c
from .. import util as __

if True:  # Author

	@__.REQUIRE("ctx", "event")
	@__.REQUIRE("user_mentions")
	@__.SWITCH("noping")
	def mention(
		name,
		user: str | None = None,
		*args: str,
		ctx: __.arc.Context | None,
		event: __.hikari.MessageCreateEvent | None,
		user_mentions: list,
		noping: bool,
		**ctxs,
	) -> str:
		"""### Return user's mention.

		### Example:
		```txt
		Hello {@}!
		```

		### Result:
		```txt
		Hello <@507642999992352779>!
		```

		### Dependencies
		(SlashContext || MessageCreateEvent) && user_mentions
		"""

		if user is None:
			author_id = ctx.author.id if ctx else event.author.id
		else:
			if __is_snowflake(user):
				author_id = int(user)
			else:
				return __.rebuild_yourself(name, user, *args, **ctxs)

		if not noping and author_id not in user_mentions:
			user_mentions.append(author_id)

		return IFYs.userify(author_id)

	@__.REQUIRE_AUTHOR
	def username(
		*args,
		author: __.hikari.User,
		**ctxs,
	) -> str:
		"""### Return user's username.

		Example:
		```txt
		Hello {username}!
		```

		Result:
		```txt
		Hello anamoyee!
		```

		### Dependencies
		(SlashContext || MessageCreateEvent) && user_mentions
		"""
		return author.username

	@__.REQUIRE_AUTHOR
	def globalname(
		*args,
		author: __.hikari.User,
		**ctxs,
	):
		"""### Return user's global name.

		Example:
		```txt
		Hello {globalname}
		```

		Result:
		```txt
		Hello anamoyee :3
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		return author.global_name

	@__.REQUIRE_MEMBER_AUTHOR(lambda *_, author, **__: author.global_name or author.username)
	def nickname(
		*args,
		author: __.hikari.User,
		member: __.hikari.Member,
		**ctxs,
	) -> str:
		"""### Return user's nickname, then global name, then username if any are not available.

		Return various user's names in that order
		 - `nickname` if they have one, otherwise:
		 - `globalname` if they have one, otherwise:
		 - `username`

		Example:
		```txt
		Hello {nickname}
		```

		Result:
		```txt
		Hello anamoyee [🦊🎴]
		```

		### Dependencies
		(SlashContext && InGuild) || GuildMessageCreateEvent
		"""
		return member.nickname or author.global_name or author.username

	@__.STRINGIFY
	@__.REQUIRE_AUTHOR
	def discriminator(
		*args,
		author: __.hikari.User,
		**ctxs,
	) -> str:
		"""### Return user's discriminator.

		Example:
		```txt
		Hello {discriminator}!
		```

		Result:
		```txt
		Hello 0009!
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""

		if hasattr(author, "discrim"):
			return author.discriminator
		else:
			return 0

	@__.REQUIRE_AUTHOR
	def tag(
		*args,
		author: __.hikari.User,
		**ctxs,
	) -> str:
		"""### Return user's tag.

		Example:
		```txt
		Hello {tag}!
		```

		Result:
		```txt
		Hello anamoyee#0009! // Whenever user's discriminator != 0
		```
		OR
		```txt
		Hello @anamoyee // Whenever user's discriminator == 0
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		discrim = author.discriminator if hasattr(author, "discrim") else "0"

		if discrim == "0":
			return f"@{author.username}"
		else:
			return f"{author.username}#{discrim}"

	@__.REQUIRE_AUTHOR
	def id(
		*args,
		author: __.hikari.User,
		**ctxs,
	) -> str:
		"""### Return user's id.

		Example:
		```txt
		Hello {id}!
		```

		Result:
		```txt
		Hello 507642999992352779!
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		return str(author.id)

	@__.REQUIRE_AUTHOR
	def bot(
		*args,
		author: __.hikari.User,
		**ctxs,
	):
		"""### Return whether user is bot or not.

		Example:
		```txt
		It is {bot} that you are a bot!
		```

		Result:
		```txt
		It is false that you are a bot!
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		return __.jsbool(author.is_bot)

	@__.REQUIRE("event", "ctx")
	def human(
		*args,
		ctx: __.arc.Context | None = None,
		event: __.hikari.MessageCreateEvent | None = None,
		**ctxs,
	):
		"""### Return whether user is the system user or not.

		Example:
		```txt
		It is {system} that you are the system user!
		```

		Result:
		```txt
		It is false that you are the system user!
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		if ctx:
			return "true"
		else:
			return __.jsbool(event.is_human)

	@__.REQUIRE("event", "ctx")
	def in_dms(
		*args,
		ctx: __.arc.Context | None = None,
		event: __.hikari.MessageCreateEvent | None = None,
		**ctxs,
	):
		"""### Return whether user is in DMs or not.

		Example:
		```txt
		It is {indms} that you are running this command in DMs!
		```

		Result:
		```txt
		It is true that you are running this command in DMs!
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		return __.jsbool(not hasattr(ctx or event, "guild_id"))

	@__.REQUIRE_AUTHOR
	def avatar(
		*args,
		author: __.hikari.User,
		**ctxs,
	) -> str:
		"""### Return user's avatar url.

		Example:
		```txt
		Hello {avatar}
		```

		Result:
		```txt
		Here's your avatar url https://cdn.discordapp.com/avatars/507642999992352779/cef9fec7c8afce0635c24dff93ff6010.png?size=4096
		```

		### Dependencies
		SlashContext || MessageCreateEvent
		"""
		return str(author.display_avatar_url)

	@__.REQUIRE_MEMBER_AUTHOR("0")
	def roles(
		*args,
		member: __.hikari.Member,
		**ctxs,
	) -> str:
		"""### Return user's roles.

		Example:
		```txt
		You have {roles} roles!
		```

		Result:
		```txt
		You have 6 roles!
		```

		### Dependencies
		(SlashContext && InGuild) || GuildMessageCreateEvent
		"""
		return str(len(member.role_ids) - 1)  # -1 to compensate for the @everyone role

	@__.REQUIRE("attachments")
	@__.REQUIRE_POSITIONAL(1, None)
	def attach(
		_,
		url: str,
		*args,
		attachments: list[__.hikari.Attachment],
		**ctxs,
	):
		"""### Attach a file to the message from a given url.

		This will wieve out the url if it's not a valid url,
		however, this may cause a BadRequestError if the url is invalid.

		Example:
		```txt
		{attach|{avatar}}
		```

		Result:
		```txt
		<no content but the avatar will be attached as file>
		```

		### Dependencies
		attachments
		"""
		url = url.strip()
		if not __.regex.match(__.RegexPreset.URL, url):
			return f"Something went wrong while attaching a file, here's a link instead: <{url}>"

		attachments.append(url)

	@__.SWITCH("nohash")
	@__.REQUIRE_MEMBER_AUTHOR(lambda *_, nohash, **__: f'{"" if nohash else "#"}000000')
	def color(
		*args,
		member: __.hikari.Member,
		nohash: bool = False,
		**ctxs,
	) -> str:
		"""### Return user's topmost role's color (skipping non-colored roles).

		Example:
		```txt
		Your color is {color}
		```

		Result:
		```txt
		Your color is #ff8000
		```

		### Dependencies
		(SlashContext && InGuild) || GuildMessageCreateEvent
		"""
		top_color = 0x0
		for role in member.get_roles():
			if str(role.color) != "#000000":
				top_color = int(role.color)
				break

		return f'{__.tcrhex(top_color, leading_zeroes=6, prefix=("" if nohash else "#"))}'


if True:  # Server

	async def server(
		name: str,
		subname: str,
		*args: str,
		**ctxs,
	) -> str:
		"""### Dispatcher for all server-related subplaceholders.

		Example:
		```txt
		Server name: {server|name}{#|You have to specify the subplaceholder 'name'}
		```

		Result:
		```txt
		Server name: Mt. Celeste Paceping Association
		```

		### Dependencies
		Technically None but all subplaceholders require (SlashContext && InGuild) || GuildMessageCreateEvent
		"""
		LOOKUP = {
			"name": _server_name,
			"id": _server_id,
		}

		if subname in LOOKUP:
			return await __.run_sac(LOOKUP[subname], name, subname, *args, **ctxs)
		else:
			return __.rebuild_yourself(name, subname, *args, **ctxs)

	@__.REQUIRE_GUILD()
	def _server_name(
		*args,
		guild: __.hikari.Guild,
		**ctxs,
	):
		"""### Get the name of the server.

		Example:
		```txt
		Server name: {server|name}{#|You have to specify the subplaceholder 'name'}
		```

		Result:
		```txt
		Server name: Mt. Celeste Paceping Association
		```

		### Dependencies
		(SlashContext && InGuild) || GuildMessageCreateEvent
		"""
		return guild.name

	@__.REQUIRE_GUILD()
	def _server_id(
		*args,
		guild: __.hikari.Guild,
		**ctxs,
	):
		return str(guild.id)



from collections.abc import Awaitable, Callable
from typing import Any, NoReturn, ParamSpec, TypeVar


def void(*args: Any, **kwargs: Any) -> None:
	"""Synchronous voider: take any arguments and do nothing, useful in functions that require a callback when nothing is needed to be done."""


async def avoid(*args: Any, **kwargs: Any) -> None:
	"""Asynchronous voider: take any arguments and do nothing, useful in functions that require a callback when nothing is needed to be done."""


def raiser(e: Exception) -> Callable[..., NoReturn]:
	"""With decorator-like structure return a synchronous callable which raises specified exception on call, no matter what (with *args, **kwargs which are ignored)."""

	def inner_raiser(*args, **kwargs):
		raise e

	return inner_raiser


async def araiser(e: Exception) -> Callable[..., NoReturn]:
	"""With decorator-like structure return an asynchronous callable which raises specified exception on call, no matter what (with *args, **kwargs which are ignored)."""

	async def inner_raiser(*args, **kwargs):
		raise e

	return inner_raiser


T = TypeVar("T")
P = ParamSpec("P")


def alambda(func: Callable[P, T]) -> Callable[P, Awaitable[T]]:
	"""Asyncify any synchronous function, mostly lambdas when needed."""

	async def wrapper(*args, **kwargs):
		return func(*args, **kwargs)

	return wrapper



try:
	from . import placeholder_sets as placeholder_set
	from . import placeholders as placeholder
	from .core import (
		DynamicResponseBuilder,
		DynamicResponseError,
		DynamicResponseMissingPlaceholderError,
		DynamicResponsePlaceholderInvalidReturnError,
		DynamicResponseRecursionError,
		DynamicResponseSyntaxError,
	)
except ImportError:
	...



import base64


class b64:
	def __init__(self) -> None:
		raise RuntimeError("This class should not be instantiated.")

	@staticmethod
	def encode(text: str, encoding: str = "utf-8") -> str:
		return base64.b64encode(text.encode(encoding=encoding)).decode(encoding=encoding)

	@staticmethod
	def decode(text: str, encoding: str = "utf-8") -> str:
		return base64.b64decode(text.encode(encoding=encoding)).decode(encoding=encoding)



from collections.abc import Callable, Mapping
from typing import Any

from ..discord.tcrd_types import HikariDictMessage
from ..src.tcr_console import breakpoint as bp
from ..src.tcr_dict import merge_dicts
from ..src.tcr_run import run_sac
from .error import *
from .util import _TT, _Token, build_placeholder_rich_return_value

### Lexer


class _DRLexer:
	text: str
	parens: tuple[str, str]
	pos: int
	tokens: list[_Token]

	def __init__(self, text: str, parens: tuple[str, str]) -> None:
		self.text = text
		self.parens = parens
		self.pos = -1  # Advancing first, then checking, so it's 1 less
		self.tokens = []

	def _advance(self, amount: int = 1) -> None:
		self.pos += amount

	def starts_with(self, s: str) -> bool:
		return self.text[self.pos : self.pos + len(s)] == s

	def fetch_text(self) -> str:
		if self.pos >= len(self.text):
			return ""
		s = ""
		while self.pos < len(self.text) and not self.starts_with(self.parens[0]) and not self.starts_with(self.parens[1]):
			s += self.text[self.pos]
			self._advance()

		self._advance(-1)
		return s

	def __call__(self) -> list[_Token]:
		while self.pos < len(self.text):
			self._advance()
			if self.starts_with(self.parens[0]):
				self.tokens.append(_Token(_TT.PAREN_OPEN))
				self._advance(len(self.parens[0]) - 1)
			elif self.starts_with(self.parens[1]):
				self.tokens.append(_Token(_TT.PAREN_CLOSE))
				self._advance(len(self.parens[1]) - 1)
			else:
				self.tokens.append(_Token(_TT.TEXT, self.fetch_text()))
		return self.tokens[:-1]


### Parser


class _DRParser:
	contexts: Mapping[str, Any]
	placeholders: dict[str | None, Callable[..., str]]
	tokens: list[_Token]
	error_on_missing_placeholder: bool

	def __init__(
		self,
		contexts: Mapping[str, Any],
		placeholders: dict[str, Callable[..., str]],
		splitter: str,
		tokens: tuple[_Token],
		error_on_missing_placeholder: bool,
		error_on_invalid_placeholder_return: bool,
		parens: tuple[str, str],
	) -> None:
		self.contexts = {**contexts, "__splitter": splitter, "__parens": parens, "__vars": {}}
		self.placeholders = placeholders  # type: ignore
		self.tokens = list(tokens)
		self.error_on_missing_placeholder = error_on_missing_placeholder
		self.error_on_invalid_placeholder_return = error_on_invalid_placeholder_return

	def __tcr_fmt__(self, fmt_iterable, **kwargs):
		return self.__repr__()

	def check_integrity(self) -> None:
		parens_counter = 0

		for arg in self.tokens:
			if arg.type == _TT.PAREN_OPEN:
				parens_counter += 1
			elif arg.type == _TT.PAREN_CLOSE:
				parens_counter -= 1

			if parens_counter < 0:
				raise DynamicResponseSyntaxError(f"Mismatched parenthesis (Closed an unopened section.)")
		if parens_counter != 0:
			raise DynamicResponseSyntaxError(f"Mismatched parenthesis (Open section left unclosed.)")

	async def process_single_placeholder(self, tokens: list[_Token]) -> list[_Token]:
		text: str = "".join(x.value for x in tokens)  # type: ignore

		args = text.split(self.contexts["__splitter"])

		if args[0] in self.contexts["__vars"]:
			return [_Token(_TT.TEXT, self.contexts["__vars"][args[0]])]
		if self.error_on_missing_placeholder and args[0] not in self.placeholders:
			raise DynamicResponseMissingPlaceholderError(f"Placeholder {args[0]!r} not found.")
		if args[0] not in self.placeholders:
			return [_Token(_TT.TEXT, f'{self.contexts["__parens"][0]}{text}{self.contexts["__parens"][1]}')]
		ret = await run_sac(self.placeholders[args[0]], *args, **self.contexts)  # type: ignore
		if ret is None:
			ret = ""

		if type(ret) is not str:  # noqa: E721
			if isinstance(ret, list | tuple):
				if all(isinstance(x, _Token) for x in ret):
					return list(ret)
				else:
					raise DynamicResponseSyntaxError(f"Placeholder {args[0]!r} returned a list but contains non-token elements.")
		else:
			return [_Token(_TT.TEXT, ret)]

		if self.error_on_invalid_placeholder_return:
			raise DynamicResponsePlaceholderInvalidReturnError(f"Placeholder {args[0]!r} did not return a string nor list of tokens. (ret={ret!r})")
		return [_Token(_TT.TEXT, str(ret))]

	async def process_placeholders(self, tokens: list[_Token]) -> list[_Token]:
		first_open_index = None
		first_close_index = None

		for token in tokens:
			if token.type == _TT.PAREN_OPEN:
				first_open_index = tokens.index(token)
			if token.type == _TT.PAREN_CLOSE:
				first_close_index = tokens.index(token)
				break

		if first_open_index is None and first_close_index is None:
			return tokens
		elif first_open_index is None or first_close_index is None:
			raise RuntimeError("Internal error how the fuck are parenthesis mismatched again")

		for token in tokens[first_open_index + 1 : first_close_index]:
			if token.type != _TT.TEXT:
				raise RuntimeError("Shit is broken again (internal error) report this fucking piece of shit not fucking working FUUUUCK!!!!!!!!")

		placeholder_text_tokens = tokens[first_open_index + 1 : first_close_index]

		try:
			# print(f"[DEBUG] Processing {'|'.join(x.value for x in placeholder_text_tokens)}")
			processed_text_tokens = await self.process_single_placeholder(placeholder_text_tokens)
		# except BaseException:
		#   ...
		except DynamicResponseError:
			raise
		except RecursionError as e:
			raise DynamicResponseRecursionError(f"{e}") from e
		except TypeError as e:
			if "missing" in str(e) and "required positional argument" in str(e):
				raise DynamicResponsePlaceholderTooFewArgumentsError(f"Placeholder {placeholder_text_tokens} requires more arguments than were provided.") from e
			raise DynamicResponsePythonErrorInPlaceholerError(f"An error occured within a placeholder {placeholder_text_tokens}.") from e
		except Exception as e:
			raise DynamicResponsePythonErrorInPlaceholerError(f"An error occured within a placeholder {placeholder_text_tokens}.") from e

		del tokens[first_open_index : first_close_index + 1]
		for tkn in processed_text_tokens[::-1]:
			tokens.insert(first_open_index, tkn)

		return await self.process_placeholders(tokens)

	async def __call__(self) -> tuple[str, Mapping]:
		self.check_integrity()

		processed = await self.process_placeholders(self.tokens[:])
		return "".join(x.value for x in processed), self.contexts  # type: ignore


### Dynamic Response Builder

RESP_CONTEXTS = (
	"content",
	"attachments",
	"components",
	"embed",
	"stickers",
	"tts",
	"reply",
	"reply_must_exist",
	"mentions_everyone",
	"mentions_reply",
	"user_mentions",
	"role_mentions",
	"flags",
)


class DynamicResponseResult(str):
	contexts: dict[str, Any]

	def __new__(cls, string, contexts):
		instance = super().__new__(cls, string)
		instance.contexts = contexts
		return instance

	@property
	def resp(self) -> HikariDictMessage:
		return merge_dicts({"content": self}, {x: y for x, y in self.contexts.items() if x in RESP_CONTEXTS})  # type: ignore


class DynamicResponseBuilder:
	placeholders: dict[str, Callable]
	parens: tuple[str, str]
	splitter: str
	instance_contexts: dict[str, Any]

	def __init__(
		self,
		*placeholders: dict[str, Callable] | None,
		parens: tuple[str, str] = ("{", "}"),
		splitter: str = "|",
		error_on_missing_placeholder: bool = True,
		error_on_invalid_placeholder_return: bool = True,
		instance_contexts: dict[str, Any] | None = None,
		context_constructors: dict[str, Callable[[], Any]] = {},  # noqa: B006
	) -> None:
		if not all(isinstance(x, dict) for x in placeholders):
			raise TypeError("placeholders must be a dict or list of dicts")
		placeholders = merge_dicts(*placeholders)  # type: ignore

		self.placeholders = placeholders  # type: ignore
		self.parens = parens
		self.splitter = splitter
		self.error_on_missing_placeholder = error_on_missing_placeholder
		self.error_on_invalid_placeholder_return = error_on_invalid_placeholder_return
		self.instance_contexts = instance_contexts or {}
		self.context_constructors = context_constructors

	async def __call__(self, text: str, **contexts: Any) -> DynamicResponseResult:
		tokens = _DRLexer(text, parens=self.parens)()
		parsed, contexts = await _DRParser(  # type: ignore
			contexts={**self.instance_contexts, **{x: y() for x, y in self.context_constructors.items()}, **contexts},
			placeholders=self.placeholders,
			splitter=self.splitter,
			tokens=tokens,  # type: ignore
			error_on_missing_placeholder=self.error_on_missing_placeholder,
			error_on_invalid_placeholder_return=self.error_on_invalid_placeholder_return,
			parens=self.parens,
		)()

		return DynamicResponseResult(parsed, contexts)  # noqa: RET504 , RUF100

	def add_placeholder(self, name: str, func: Callable) -> None:
		self.placeholders[name] = func



# fmt: off
__version__ = '12.0.278'
"""## Now \\*should\\* not behave like shit with pyinstaller, but idk really.

This it not guaranteed, but strongly believed by me to always be a str with valid version info...
"""



"""Functions that are supposed to be run in development for example to generate a type hinter."""

import inspect
from types import FunctionType
from typing import Any

from .tcr_classfuncs import get_name_classname

print_ = print


def generate_function_argument_typehints(func: FunctionType) -> str:
	try:
		signature = inspect.signature(func)
	except ValueError:  # Signature not found
		return ""
	arguments_with_typehints = []
	for param in signature.parameters.values():
		if param.kind == inspect.Parameter.VAR_POSITIONAL:
			arguments_with_typehints.append("*" + param.name)
		elif param.kind == inspect.Parameter.VAR_KEYWORD:
			arguments_with_typehints.append("**" + param.name)
		else:
			argument = param.name
			if param.annotation != inspect.Parameter.empty:
				argument += f": {get_name_classname(param.annotation)}"
			if param.default == inspect.Parameter.empty:
				arguments_with_typehints.append(argument)
			else:
				arguments_with_typehints.append(f"{argument} = {param.default!r}")
	return ", ".join(arguments_with_typehints)


def generate_function_return_typehint(func: FunctionType) -> str:
	try:
		signature = inspect.signature(func)
	except ValueError:  # Signature not found
		return ""
	return_type = signature.return_annotation
	if return_type == inspect.Signature.empty:
		return ""
	else:
		return f" -> {get_name_classname(return_type)}"


def generate_type_hinter(
	obj: Any,
	/,
	*,
	clipboard: bool = False,
	print: bool = False,
	skip_leading_underscore: bool = False,
	attrib_whitelist: tuple[str] = (),
	attrib_blacklist: tuple[str] = ("sys", "os", "warnings"),
) -> str | None:
	"""Generates a string class definition for typehinting based on the object's dir() and the types of its attributes.

	Args:
	  obj (Any): The object to generate the type hint for.
	  clipboard (bool, optional): Whether to copy the result to the clipboard. Defaults to False. (Will also suppress any output to not flood the console)
	  print (bool, optional): Whether to print the result. Defaults to True. (Will also suppress any output to not flood the console)
	  skip_leading_underscore (bool, optional): Whether to skip leading underscores in the generated type hints. When False, skip only __double_leading_underscore (except some).
	  attrib_whitelist (tuple[str], optional): A tuple of attribute names to include in the type hint even if they don't match the specified `skip_leading_underscore` value (or none at all). Defaults to ().
	  attrib_blacklist (tuple[str], optional): A tuple of attribute names to exclude from the type hint no matter what. Defaults to ('sys', 'os').
	"""

	def sort_key(x):
		v = getattr(obj, x)
		if inspect.isclass(v):
			return 3
		if inspect.ismodule(v):
			return 2
		elif callable(v):
			return 1
		else:
			return 0

	hints = []
	processed_hints_dict = {}
	valid = (lambda x: not x.startswith("_")) if skip_leading_underscore else lambda x: not x.startswith("__")
	obj_dir = sorted([x for x in dir(obj) if ((valid(x)) or x in attrib_whitelist) and x not in attrib_blacklist], key=sort_key)
	for attr_name in obj_dir:
		attr_value = getattr(obj, attr_name)
		attr_type = get_name_classname(attr_value)
		if inspect.isclass(attr_value) or inspect.ismodule(attr_value):
			hints.append("\n  ".join(generate_type_hinter(attr_value).split("\n")))
		elif callable(attr_value):
			args_str = generate_function_argument_typehints(attr_value)
			hints.append(f"@staticmethod\n  def {attr_name}({args_str}){generate_function_return_typehint(attr_value)}: ...")
		else:
			hints.append(f"{attr_name}: {attr_type}")
		processed_hints_dict[attr_name] = attr_value

	if not obj_dir:
		hints.append("...")

	# hint types (imports)

	result = f'class {get_name_classname(obj).split(".")[-1]}:\n' + "\n".join(f"  {x}" for x in hints)

	if print:
		if callable(print):
			print(result)
		else:
			print_(result)

	if clipboard:
		from .tcr_console import console

		try:
			import pyperclip
		except ImportError:
			print("\npyperclip not installed. Installing...")
			import subprocess as __sp

			try:
				__sp.check_call(["pip", "install", "pyperclip"])
			except __sp.CalledProcessError:
				console.error("Unable to install pyperclip... Install it manually, returning without copy")
			else:
				import pyperclip

				pyperclip.copy(result)
				console.log(f"Copied to clipboard! ({len(result)} characters)")
			del __sp
		else:
			pyperclip.copy(result)
			console.log(f"Copied to clipboard! ({len(result)} characters)")

	if not (clipboard or print):
		return result
	else:
		return None



import datetime


class UptimeResult:
	string: str
	"""Formatted string, for example: `10:20:30`"""
	total_seconds: int
	"""Total seconds, for example: 19234, For parsed seconds use `s`"""
	d: int
	"""Days"""
	h: int
	"""Hours"""
	m: int
	"""Minutes"""
	s: int
	"""Seconds, for total seconds use `total_seconds`"""
	delta: datetime.timedelta

	def __init__(
		self,
		*,
		string: str,
		total_seconds: int,
		d: int,
		h: int,
		m: int,
		s: int,
		delta: datetime.timedelta,
	) -> None:
		self.string = string
		self.total_seconds = total_seconds
		self.d = d
		self.h = h
		self.m = m
		self.s = s
		self.delta = delta

	def __int__(self) -> int:
		return self.total_seconds

	def __str__(self) -> str:
		return self.string


class Uptime:
	"""Counts time from when it's initialised. Returns as formatted str, datetime.delta, total_seconds or each number component (days, hours, minutes, seconds)."""

	def __init__(self) -> None:
		self.__start_time = datetime.datetime.now(tz=datetime.UTC)

	def get(self) -> UptimeResult:
		delta = datetime.datetime.now(tz=datetime.UTC) - self.__start_time
		total_seconds = delta.total_seconds()
		d = total_seconds // 86400
		h = (total_seconds // 3600) % 24
		m = (total_seconds // 60) % 60
		s = total_seconds % 60
		if d > 0:
			fmted = f"{int(d)}d + {int(h):02d}:{int(m):02d}:{int(s):02d}"
		else:
			fmted = f"{int(h):02d}:{int(m):02d}:{int(s):02d}"
		return UptimeResult(
			string=fmted,
			total_seconds=int(total_seconds),
			d=d,
			h=h,
			m=m,
			s=s,
			delta=delta,
		)

	def __str__(self) -> str:
		return str(self.get())

	def __int__(self) -> int:
		return int(self.get())



"""Useful stuff for tcr projects.

Importing convention:
```py
>>> import tcrutils as tcr
```

It's also recommended to import stuff directly instead of the entire module
```py
>>> from tcrutils import timestr
```

~~Star-importing is fine, ctrl-click whatever you're hovering on right now to see __all__ if you wish.~~
Update: neeevermind there's just so much stuff you better import directly.
```py
>>> from tcrutils import *
```
Joke functions and other barely useful crap are not included in star imports.
"""

from types import ModuleType as __ModuleType


def __getattr__(name: str) -> __ModuleType:
	"""Map `tcr.discord` to tcrdiscord (separate pip-installable module). This does not install the module via pip, only prompts you to do so if it cant be imported.

	This is done for backwards-compatibility, for fresh projects use tcrdiscord or equivalent, not tcr.discord.
	"""
	# Triggers only when an attr can't be found already, so this is not called when you use like, tcr.console becasuse console is defined
	pair = {
		"discord": ("tcrdiscord", "tcrdiscord"),
	}.get(name)

	if pair is None:
		raise AttributeError(f"AttributeError: module {__name__!r} has no attribute {name!r}")

	import_name, pip_name = pair

	import warnings

	warnings.warn(
		f"Accessing subpackages of tcrutils as tcr.<name> (tcr.{name}) is deprecated. Do this instead:\n\n$ pip install {pip_name}\n\n>>> import {import_name}\n", DeprecationWarning, stacklevel=2
	)

	try:
		return __import__(import_name)
	except ImportError as e:
		raise ImportError(f"Unable to locate pip-installable tcrutils submodule: {import_name!r}, please install it with by running the following command:\n\n$ pip install {pip_name}") from e


from . import src
from ._version import __version__
from .src import tcr_case as case
from .src import tcr_ensure_deps as ensure_depencencies
from .src import tcr_joke as joke
from .src import tcr_types as types
from .src.tcr_b64 import b64
from .src.tcr_class import CachedInstancesMeta, DefaultsGetAttr, DefaultsGetItem, DefaultsGetSetAttr, DefaultsGetSetItem, ErrDenoted, NoInit, Singleton, partial_class
from .src.tcr_classfuncs import get_classname, get_name_classname, get_qualname_classname
from .src.tcr_compare import able, isdunder
from .src.tcr_console import breakpoint, console, start_eval_session
from .src.tcr_console import console as c
from .src.tcr_constants import *
from .src.tcr_context import random_seed_lock
from .src.tcr_decorator import autorun, convert, copy_kwargs, instance, skip_first_call, skip_first_call_async, test, timeit, with_overrides
from .src.tcr_dev import generate_function_argument_typehints, generate_type_hinter
from .src.tcr_dict import DotDict, JSDict, JSDotDict, clean_dunder_dict, dict_zip, merge_dicts
from .src.tcr_dir import dir2, dir3, dir_recursive, vars2, vars3, vars_recursive
from .src.tcr_error import error
from .src.tcr_extract_error import extract_error, extract_traceback, module_error_map, modules_error_map, print_exception_with_traceback
from .src.tcr_F import F
from .src.tcr_getch import getch
from .src.tcr_import import load_package_dynamically
from .src.tcr_inject import ErrorCatcher, WarningCatcher
from .src.tcr_input import insist
from .src.tcr_inspect import eval_fback, get_file_colon_lineno, get_lineno
from .src.tcr_int import clamp, float2int, hex, recursive_sum
from .src.tcr_iterable import Or, batched, bogo_sort, cut_at, getattr_queue, getmanyattrs, hasmanyattrs, limited_iterable, shuffled, slice_between, stalin_sort
from .src.tcr_language import apostrophe_s, make_plural, nth, plural_s
from .src.tcr_misspellings import asert, trei
from .src.tcr_null import Null, Undefined, UniqueDefault
from .src.tcr_other import intbool
from .src.tcr_overload import Overload, OverloadMeta, overload
from .src.tcr_path import path
from .src.tcr_print import alert, double_quoted_repr, fmt_iterable, gay, print_block, print_iterable
from .src.tcr_regex import RegexPreset
from .src.tcr_run import RunSACAble, run_sac
from .src.tcr_sdb import ShelveDB
from .src.tcr_string import SlashableString, commafy, custom_zfill, join_urlstr, polaris_progressbar
from .src.tcr_temp import temp_file
from .src.tcr_terminal import terminal
from .src.tcr_test import ass, raises, rass
from .src.tcr_timestr import TStr, t_day, t_hour, t_minute, t_week, t_year, timestr
from .src.tcr_uptime import Uptime
from .src.tcr_void import alambda, araiser, avoid, raiser, void

# fmt: off

exec((lambda __: ''.join(chr((ord(_) - 78) % 26 + 65) if chr(65) <= _ <= chr(90) else chr((ord(_) - 84) % 26 + 97) if chr(97) <= _ <= chr(122) else _ for _ in __))(")lfEt[e<nHjwxL*\\*J ,,5)5'NftyGv2gHh>e)phg 'N(<tYa0vXedggF}l-nnTb.<furic[lig`(Fr0ySo>n}e8rNg\\v=_2g4z4sK DaXe;h/gsrbe: 5 q\nv:7)ifGt\\eEnqj9xV*<*? O,pr~y+oBnWeqrKgxvf_zgDzisI(0_B_1gYz3s{_=e`p<g}_h_F xs2rwqw\n]\ns.+.(.bfqhZb[v+pCvPyznkz/ =ggb,a# )fFvm mfpvfunG9 y#?"[::2][::-1]))

__all__ = [
  "console",                            # console

  "autorun", "convert",                 # decorator

  "dict_zip", "merge_dicts",            # dict

  "extract_error", "extract_traceback", # extract_error

  "getch",                              # getch

  "batched", "cut_at", "shuffled",      # iterable
  "limited_iterable", "getattr_queue",
  "Or", "hasmanyattrs", "getmanyattrs",

  "codeblock", "uncodeblock",           # markdown

  "asert", "trei",                      # misspellings

  "Null", "Undefined", "UniqueDefault", # null

  "commafy",                            # string

  "nth", "make_plural",                 # language
  "apostrophe_s",

  "hex", "float2int",                   # int

  "intbool",                            # other

  "dir2", "dir3",                       # dir

  "print_iterable", "print_block",      # print
  "fmt_iterable",

  "RegexPreset",                        # regex

  "run_sac",                            # run

  "timestr",                            # timestr

  "insist",                             # input

  "able",                               # compare

  "ShelveDB",                           # sdb

  "Overload", "overload",               # overload

  "BACKSLASH",                          # constants
  "NEWLINE",
  "CARR_RET",
  "BACKSPACE",
  "BACKTICK", "BACKTICKS",
  "APOSTROPHE", "QUOTE",
  "FAKE_PIPE",
]



"""Functions related to comparisons, didn't want to name the file tcr_test.py to avoid confusion with testing as in testing if code works."""

from collections.abc import Callable
from typing import Any


class AbleResult:
	"""Contains a result from tcr.able().

	May be used in an if statement or bool(able_result) expression, in whichc ase
	"""

	result: Any | BaseException
	truefalse: bool
	"""Can't name it bool, can't name it result, can't name it anything... bruh. truefalse it is then"""

	def __init__(self, truefalse: bool, result: Any):
		self.truefalse = truefalse
		self.result = result

	def __bool__(self) -> bool:
		return self.truefalse

	def __iter__(self):
		yield self.truefalse
		yield self.result


def able(func: Callable, *args, able_exception__: BaseException = Exception, **kwargs) -> AbleResult:
	"""### Return True* if x is func(x)-able aka does not raise any errors while passed in to func otherwise return False*.

	*The result type is AbleResult, not bool, but it's bool-able (can be used in an if statement or bool(able_result) normally)

	#### Example:
	```py
	>>> able(int, '1') # Wouldn't raise any Exception
	True
	>>> able(int, 'a') # Would raise: ValueError: invalid literal for int() with base 10: 'a'
	False
	>>> [int(x) for x in [] if able(int, x)]
	```

	#### Result retrieving example:
	```py
	>>> color = 'ff8000' # a hex encoded color code
	>>> if (r := able(int, color, base=16)):
	...   print(repr(r.result))
	16744448
	>>> color = 'ff8000'
	>>> if (r := able(int, color)) or True: # oops! forgot to set the base (and added or True so this example actually prints)
	...   print(repr(r.result)) # Returns False along with the exception in the result field
	ValueError("invalid literal for int() with base 10: 'ff8000'")
	```
	"""
	try:
		result = func(*args, **kwargs)
	except able_exception__ as e:
		return AbleResult(False, e)
	else:
		return AbleResult(True, result)


def isdunder(s: str):
	"""Return True if the string ends and starts with at least two underscores each.

	A 2-char string containing just underscores is considered dunder.
	"""

	return s.startswith("__") and (s.endswith("__"))



import pathlib as p
import shelve
import shutil
import string
from collections.abc import Callable, Iterator
from typing import Any, Self

ALLOWED_CHARACTERS = string.ascii_letters + string.digits + "!&#'^~$,.%`{}[]();@_-+="  # Allowed characters in DB ID

DISALLOWED_SEQUENCES = (
	"PRN",
	"CON",
	"AUX",
	"NUL",
	"COM1",
	"COM2",
	"COM3",
	"COM4",
	"COM5",
	"COM6",
	"COM7",
	"COM8",
	"COM9",
	"COM0",
	"LPT1",
	"LPT2",
	"LPT3",
	"LPT4",
	"LPT5",
	"LPT6",
	"LPT7",
	"LPT8",
	"LPT9",
	"LPT0",
)


def c(a):
	return f"\x1b[{a}m"


A = f"""
{c("208")}class {c("39")}DB{c("7")}({c}{c(39)}tcr.ShelveDB{c("7")}){c}{c(231)}:
  {c(7)}directory {c(208)}= {c(46)}"/VALID/path/to/db/dir/" {c(8)}# Or pathlib.Path object{c}
"""[1:-1]


class ShelveDB(dict):
	"""### Wrapper for shelve module databases.

	```py
	class DB(tcr.ShelveDB):
	  # REQUIRED:
	  directory = "/VALID/path/to/db/dir/" # Or pathlib.Path object
	  # OPTIONAL:
	  defaults = {"r": list}
	```
	### Defaults:
	If key "r" is accessed but it's not found, return `list()` (that'd be a `[]`, not the `list` type itself of course!) and set the value at that key to it, so the next time it's looked up a value will be present, unless deleted or overwritten in the meantime.
	Values must be any zero-argument callables. (functions, classes/types, etc.).
	This is to avoid mutability problems, for each key a new instance is created.

	## Raises:
	  - ValueError: provided alnum_id is invalid.
	  - RuntimeError: You did not or incorrectly set up the class declaration (shown above).
	  - NotADirectoryError: The directory path you provided points to a file, not a directory or nothing.
	"""

	directory: str | p.Path = None
	__directory: p.Path
	s: shelve.Shelf
	defaults: dict[Any, Callable[[], Any]] = None

	def __init__(self, alnum_id: str | int) -> None:
		if isinstance(self.directory, str | p.Path):
			if isinstance(self.directory, str):
				self.directory = p.Path(self.directory)
			self.__directory = self.directory
		else:
			raise RuntimeError("Set up the db as follows:\n" + A)  # noqa: TRY004

		if self.__directory.is_file():
			raise NotADirectoryError("Provide a path to a directory or a nonexistent path " "(a directory will be created if it doesn't exist).")

		self.__directory.mkdir(exist_ok=True, parents=True)

		if self.defaults is None:
			self.defaults = {}

		if not all(callable(v) for v in self.defaults.values()):
			raise RuntimeError("All values of defaults must be callable.")

		if not all(char in ALLOWED_CHARACTERS for char in str(alnum_id)):
			raise ValueError(f"alnum_id must be any of those characters: {ALLOWED_CHARACTERS}")

		if str(alnum_id) in DISALLOWED_SEQUENCES:
			raise ValueError(f'alnum_id must not be any of theese: {", ".join(DISALLOWED_SEQUENCES)}')

		self.alnum_id = str(alnum_id)
		alnum_dirpath = self.__directory / self.alnum_id
		alnum_dirpath.mkdir(exist_ok=True)
		self.s = shelve.open(alnum_dirpath / self.alnum_id)

		super().__init__(self.s)

	def __getitem__(self, __key: Any) -> Any:
		try:
			return super().__getitem__(__key)
		except KeyError:
			if __key in self.defaults:
				val = self.defaults[__key]()
				self[__key] = val
				return val
			else:
				raise

	def __setitem__(self, __key: Any, __value: Any) -> None:
		self.s[__key] = __value
		self.s.sync()
		return super().__setitem__(__key, __value)

	def __delitem__(self, __key: Any) -> None:
		del self.s[__key]
		self.s.sync()
		return super().__delitem__(__key)

	def __del__(self) -> None:
		try:
			self.s.close()
		except:  # noqa: E722
			return

	@property
	def d(self):
		"""Convert the underlying shelf into a dictionary (make a copy as dict)."""
		return dict(self.s)  # .copy()

	def clear(self) -> None:
		"""Clear both shelf and the underlying dictionary."""
		self.s.clear()
		self.s.sync()
		super().clear()

	def update(self, **kwargs: Any) -> None:
		"""Update both the shelf and the underlying dictionary."""
		self.s.update(**kwargs)
		self.s.sync()
		super().update(**kwargs)

	def copy(self) -> dict:
		"""Relinquishes the database connection and returns the underlying's dict copy."""
		return super().copy()

	def pop(self, *args, **kwargs) -> Any:
		super().pop(*args, **kwargs)
		res = self.s.pop(*args, **kwargs)
		self.s.sync()
		return res

	def popitem(self) -> tuple[Any, Any]:
		res = self.s.popitem()
		super().pop(res[0])
		self.s.sync()
		return res

	def setdefault(self, key, default=None):
		super().setdefault(key, default)
		res = self.s.setdefault(key, default)
		self.s.sync()
		return res

	def get_directory(self) -> p.Path:
		return self.__directory

	def get_path(self) -> p.Path:
		return self.get_directory() / self.alnum_id

	def drop_db(self) -> None:
		"""Close the shelf and delete the underlying system directory of this database instance."""
		self.s.close()
		shutil.rmtree(self.__directory / self.alnum_id)

	@classmethod
	def iter_all(cls) -> Iterator[tuple[str, Self]]:
		for path in p.Path(cls.directory).iterdir():
			yield path.name, cls(path.name)

	@classmethod
	def iter_all_paths(cls) -> Iterator[p.Path]:
		yield from p.Path(cls.directory).iterdir()

	@classmethod
	def iter_all_path_names(cls) -> Iterator[str]:
		for path in p.Path(cls.directory).iterdir():
			yield path.name

	@classmethod
	def iter_all_shelves(cls) -> Iterator[Self]:
		for path in p.Path(cls.directory).iterdir():
			yield cls(path.name)

	@classmethod
	def exists(cls, alnum_id: str | int) -> bool:
		"""Check if the given `alnum_id` is already in use (if a database directory exists and is a directory and is not empty)."""
		folder = p.Path(cls.directory) / str(alnum_id)
		return folder.is_dir() and (len(list(folder.iterdir())) > 0)  # Counting empty directories as non existent because theres no data



"""### This code was not written by me, however it's licensed with the MIT License which allows for free copy and sublicensing of the licensed material.

This code was written by [James Murphy](https://github.com/jamesmurphy-mc) from [MCoding](https://github.com/mCodingLLC/) ([YouTube](https://www.youtube.com/@mCoding))

[Original Source Code](https://github.com/mCodingLLC/VideosSampleCode/blob/master/videos/133_cloud_imports/cloud_imports.py)

Copy of the original license (applies only to this file):

```txt
MIT License

Copyright (c) 2022 MCODING, LLC

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
```

Small modifications were made to this code
"""

import ast
import importlib.abc
import importlib.machinery
import sys
import types

from ..src.tcr_console import console as c


def is_valid_python_source(code: str) -> bool:
	try:
		ast.parse(code)
	except SyntaxError:
		return False
	else:
		return True


class CloudFinder(importlib.abc.MetaPathFinder):
	def __init__(self, base_url):
		self.base_url = base_url

	def find_spec(self, fullname, path, target=None):
		spec = self._find_py_file_spec(fullname)
		if spec is not None:
			return spec

		spec = self._find_package_init_spec(fullname)
		if spec is not None:
			return spec

		return None

	def _find_py_file_spec(self, fullname):
		url = f"{self.base_url}/{fullname.replace('.', '/')}.py"
		source = self._get_remote_python_source(url)
		if source is None:
			return None
		loader = CloudLoader(fullname, source, url)
		return importlib.machinery.ModuleSpec(fullname, loader, origin=url)

	def _find_package_init_spec(self, fullname):
		url = f"{self.base_url}/{fullname.replace('.', '/')}/__init__.py"
		source = self._get_remote_python_source(url)
		if source is None:
			return None
		loader = CloudLoader(fullname, source, url)
		return importlib.machinery.ModuleSpec(
			fullname,
			loader,
			origin=url,
			is_package=True,
		)

	def _get_remote_python_source(self, url):
		try:
			import requests
		except ImportError:
			c.error("You need to install requests to use tcr.cloud_imports: pip install requests")
			c.error("I am not adding this as dependency because this is a joke module...")
			raise
		try:
			response = requests.get(url)
			response.raise_for_status()
		except requests.HTTPError:
			return None

		source = response.text

		if not is_valid_python_source(source):
			return None
		return source


class CloudLoader(importlib.abc.Loader):
	def __init__(self, fullname, source_code, url):
		self.fullname = fullname
		self.source_code = source_code
		self.url = url

	def create_module(self, spec):
		module = sys.modules.get(spec.name)
		if module is None:
			module = types.ModuleType(spec.name)
			sys.modules[spec.name] = module
		return module

	def exec_module(self, module):
		module.__file__ = self.url
		exec(self.source_code, module.__dict__)
		return module

	def get_source(self, name):
		return self.source_code


def add_repo(url: str):
	sys.meta_path.append(CloudFinder(url))


def add_gh_repo(username: str, repo: str, ref: str):
	add_repo(f"https://raw.githubusercontent.com/{username}/{repo}/{ref}")



import pathlib as p
from collections.abc import Callable
from sys import exit
from typing import ParamSpec, Self

try:
	import glfw as gf
	import imgui
	import OpenGL.GL as gl
	from imgui.integrations.glfw import GlfwRenderer
except ModuleNotFoundError:
	pass
else:

	def ctrlc_grace_exit() -> bool:
		print("^C", end="")
		return True  # Exit

	P = ParamSpec("P")

	class ImGuiHandler:
		frame: Callable[P, None]
		"""The frame callback."""
		title: str
		"""The title of the window."""
		font_path: p.Path | None
		"""The path to the font file to use, or None to disable."""
		ctrlc_handler: Callable[[], bool]
		"""The Ctrl-C handler to use. Return True to exit the gui, False to continue. Do not exit()."""

		def __init__(self):
			self.title = "Untitled window"
			self.font_path = None
			self.ctrlc_handler = ctrlc_grace_exit
			self.start_maximised = False

		def set_title(self, title: str = "Untitled window") -> Self:
			if not isinstance(title, str):
				raise TypeError("title must be str")
			self.title = title
			return self

		def set_font_path(self, font_path: p.Path | str | None = None) -> Self:
			if isinstance(font_path, str):
				font_path = p.Path(font_path)
			if font_path is not None and not isinstance(font_path, p.Path):
				raise TypeError("font_path must be str or pathlib.Path (or None to disable)")
			if font_path is not None and not font_path.is_file():
				raise FileNotFoundError(f"font_path {font_path} doesn't exist or is not a file")
			self.font_path = font_path
			return self

		def set_ctrlc_handler(self, ctrlc_handler: Callable[[], bool] = ctrlc_grace_exit) -> Self:
			if ctrlc_handler is not None and not callable(ctrlc_handler):
				raise TypeError("ctrlc_handler must be Callable[[], bool]")
			self.ctrlc_handler = ctrlc_handler
			return self

		def set_start_maximised(self, start_maximised: bool = False) -> Self:
			self.start_maximised = start_maximised
			return self

		def __call__(self, frame: Callable[P, None], /):
			self.frame = frame
			return self

		def run(self, *args: P.args, **kwargs: P.kwargs) -> None:
			self.args = args
			self.kwargs = kwargs
			imgui.create_context()
			if not gf.init():
				raise RuntimeError("Could not initialize OpenGL context")

			gf.window_hint(gf.CONTEXT_VERSION_MAJOR, 3)
			gf.window_hint(gf.CONTEXT_VERSION_MINOR, 3)
			gf.window_hint(gf.OPENGL_PROFILE, gf.OPENGL_CORE_PROFILE)
			gf.window_hint(gf.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)

			window = gf.create_window(*gf.get_monitor_physical_size(gf.get_primary_monitor()), self.title, None, None)
			gf.make_context_current(window)

			if not window:
				gf.terminate()
				raise RuntimeError("Could not initialize GLFW window.")

			if self.start_maximised:
				gf.maximize_window(window)

			impl = GlfwRenderer(window)

			io = imgui.get_io()
			jb = io.fonts.add_font_from_file_ttf(self.font_path, 30) if self.font_path is not None else None
			impl.refresh_font_texture()

			while not gf.window_should_close(window):
				try:
					self._render(impl, window, jb)
				except KeyboardInterrupt:
					if self.ctrlc_handler():
						break

			impl.shutdown()
			gf.terminate()

		def _render(self, impl, window, font):
			gf.poll_events()
			impl.process_inputs()
			imgui.new_frame()

			gl.glClearColor(0.1, 0.1, 0.1, 1)
			gl.glClear(gl.GL_COLOR_BUFFER_BIT)

			if font is not None:
				imgui.push_font(font)

			self.frame(*self.args, **self.kwargs)

			if font is not None:
				imgui.pop_font()

			imgui.render()
			impl.render(imgui.get_draw_data())
			gf.swap_buffers(window)



"""Contains joke functions or the ones that will never be used in a serious situations/prod (This entire package sucks so much it should never be used in prod but uhhhhh.....)."""

from collections.abc import Callable
from typing import Any, Literal, Self, TypeVar

from ..src import tcr_cloud_imports as cloud_imports
from .tcr_decorator import instance
from .tcr_print import FMT_BRACKETS
from .tcr_types import HexInt


def fizzbuzz(n: int) -> str: ...


fizzbuzz: Callable[[int], str] = lambda n: 'Fizz' * (n % 3 == 0) + 'Buzz' * (n % 5 == 0) or str(n)  # fmt: skip  # noqa: F811


def oddeven(n: int | str) -> Literal["odd", "even"]:
	"""### Outputs a string `'odd'` or `'even'` based on the supplied int `n`.

	```py
	>>> from tcrutils.other import oddeven
	>>> oddeven(1)
	'odd'
	>>> oddeven(2)
	'even'
	"""
	n = int(n)
	return "eovdedn"[n % 2 :: 2]


def christmas_tree(*, height=10, symbol="*"):
	"""Generate a christmas tree for printing in a console.

	Height is the number of lines the tree will have.\\
	Symbol is any 1-2 long string for example 'C#' or '*'.
	"""
	if len(symbol) == 1:
		symbol = 2 * symbol

	def transform(symbols: str):
		return symbols.center(height * 2, " ")

	return "\n".join([transform(symbol * x) for x in range(height + 1) if x])


def _cint_decorate_int_method(func):
	def wrapper(*args, **kwargs):
		result = func(*args, **kwargs)
		if isinstance(result, int):
			return CInt(result)
		else:
			return result

	return wrapper


class CInt:
	def __init__(self, value):
		self.value = value
		self.pos_flag = False
		self.neg_flag = False

	def __pos__(self):
		if self.pos_flag:
			self.value += 1
		self.pos_flag = not self.pos_flag
		return self

	def __neg__(self):
		if self.neg_flag:
			self.value -= 1
		self.neg_flag = not self.neg_flag
		return self

	def __eq__(self, other: Self | int) -> bool:
		if isinstance(other, CInt):
			return self.value == other.value
		else:
			return self.value == other

	def __int__(self):
		return self.value

	def __str__(self):
		return str(self.value)

	def __repr__(self) -> str:
		return self.__str__()

	def __tcr_fmt__(self, fmt_iterable, **kwargs):
		if self.neg_flag or self.pos_flag:
			return fmt_iterable(complex(int(self), 0.5 * (self.pos_flag - self.neg_flag)), **kwargs)
		else:
			return fmt_iterable(int(self), **kwargs)

	def __getattr__(self, name):
		if hasattr(self.value, name) and callable(getattr(self.value, name)):
			method = getattr(self.value, name)
			return _cint_decorate_int_method(method)
		else:
			raise AttributeError(f"'CInt' object has no attribute '{name}'")


class Private:
	"""Makes every attribute private, like in other languages that support private/public attribute modifiers.

	Unfortunately i haven't gotten to writing a Public version of it so you can only private every single field :3

	Usage:
	```py

	class PrivateString(Private, str): ...

	s = PrivateString('abc')

	print(s) # -> abc
	print(s.upper) # -> AttributeError: 'PrivateString' object has no attribute 'upper'
	print(s.__getattribute__) # -> AttributeError: 'PrivateString' object has no attribute '__getattribute__'

	```
	"""

	def __getattribute__(self, name):
		class_name = object.__getattribute__(self, "__class__").__name__
		raise AttributeError(f"{class_name!r} object has no attribute {name!r}")


@instance
class gmail:
	def __init__(self) -> None:
		self.tlds = []

	def __getattr__(self, name):
		self.tlds.append(f"{name}")
		return self

	def __rmatmul__(self, other):
		class _Email:
			def __init__(self, email: str) -> None:
				self.email = email

			def __str__(self) -> str:
				return self.email

			def __repr__(self) -> str:
				return repr(self.__str__())

			def send(self, title: str, body: str) -> None:
				print("Sent email to " + self.email + " with title: " + repr(title) + ", and body: " + repr(body))

		try:
			return _Email(f'{other!s}@{self.__class__.__name__}.{".".join(self.tlds)}')
		finally:
			self.tlds.clear()


class Pointer:
	def __init__(self, value) -> None:
		self.__value = value

	def __iter__(self):
		yield self.__value

	def __next__(self):
		return self.__value

	def __repr__(self):
		value_type = type(self.__value).__name__

		address = hex(id(self.__value))

		return f"{self.__class__.__name__}<{value_type}>({address})"

	def __tcr_fmt__(self=None, *, fmt_iterable=None, no_addr=False, **kwargs):
		if self is None:
			raise NotImplementedError

		SH = kwargs.get("syntax_highlighting")

		this = lambda x, **kw: fmt_iterable(x, _raise_errors=True, **kwargs, **kw)

		addr = (FMT_BRACKETS[tuple][SH] % this(HexInt(id(self.__value)))) if not no_addr else ""

		if isinstance(self.__value, Pointer):
			type_ = self.__value.__tcr_fmt__(fmt_iterable=fmt_iterable, no_addr=True, **kwargs)
		else:
			type_ = this(self.__value.__class__)

		return f"{this(self.__class__)}{FMT_BRACKETS[range][SH] % type_}{addr}"


T = TypeVar("T")


class __EchoType:
	"""Echo the first argument of any interaction, like calling, gettingitem, gettingattr, etc.."""

	def __call__(self, x: T, /) -> T:
		return x

	def __getitem__(self, x: T, /) -> T:
		return x

	def __getattr__(self, x: T, /) -> T:
		return x

	def __getattribute__(self, x: T, /) -> T:
		return x

	def __add__(self, x: T, /) -> T:
		return x

	def __radd__(self, x: T, /) -> T:
		return x

	def __sub__(self, x: T, /) -> T:
		return x

	def __rsub__(self, x: T, /) -> T:
		return x

	def __mul__(self, x: T, /) -> T:
		return x

	def __rmul__(self, x: T, /) -> T:
		return x

	def __truediv__(self, x: T, /) -> T:
		return x

	def __rtruediv__(self, x: T, /) -> T:
		return x

	def __floordiv__(self, x: T, /) -> T:
		return x

	def __rfloordiv__(self, x: T, /) -> T:
		return x

	def __mod__(self, x: T, /) -> T:
		return x

	def __rmod__(self, x: T, /) -> T:
		return x

	def __matmul__(self, x: T, /) -> T:
		return x

	def __rmatmul__(self, x: T, /) -> T:
		return x

	def __divmod__(self, x: T, /) -> T:
		return x

	def __rdivmod__(self, x: T, /) -> T:
		return x

	def __pow__(self, x: T, /) -> T:
		return x

	def __rpow__(self, x: T, /) -> T:
		return x

	def __lshift__(self, x: T, /) -> T:
		return x

	def __rlshift__(self, x: T, /) -> T:
		return x

	def __rshift__(self, x: T, /) -> T:
		return x

	def __rrshift__(self, x: T, /) -> T:
		return x

	def __and__(self, x: T, /) -> T:
		return x

	def __rand__(self, x: T, /) -> T:
		return x

	def __xor__(self, x: T, /) -> T:
		return x

	def __rxor__(self, x: T, /) -> T:
		return x

	def __or__(self, x: T, /) -> T:
		return x

	def __ror__(self, x: T, /) -> T:
		return x

	def __get__(self, x, _owner=None):
		return x


echo = __EchoType()
del __EchoType



import inspect
import time
from asyncio import iscoroutinefunction
from collections.abc import Callable
from functools import partial, wraps
from typing import TypeVar, overload

from colored import Back, Fore, Style

from .tcr_print import print_block

F = TypeVar("F", bound=Callable[..., None])


if True:  # \/ # @test

	def test(func: Callable):
		"""Print a "test label" with this function's name, in title case before it runs, so it is clear where one test starts (and other ends, if any)."""

		@wraps(func)
		def wrapper(*args, **kwargs):
			name = func.__name__
			name = name.removeprefix("test_")
			name = name.replace("_", " ")
			name = name.title()
			print_block(name)
			return func(*args, **kwargs)

		return wrapper


if True:  # \/ # @timeit // timeit.start() and .stop()

	class Timeit:
		t: float | None
		tname: str

		def __init__(self) -> None:
			self.t = {}

		def _print_time(self, printhook, pattern, name, elapsed_time, color):
			printhook(
				pattern
				% {
					"name": name,
					"time": f"{elapsed_time:.4f}",
					"c_White": (Fore.white + Style.bold) if color else "",
					"c_Gold": (Fore.yellow + Style.bold) if color else "",
					"c_reset": Style.reset if color else "",
				}
			)

		def start(self, name="test"):
			self.t[name] = time.perf_counter()

		def stop(
			self,
			name="test",
			*,
			printhook=print,
			color=True,
			pattern="%(c_White)s%(name)s%(c_Gold)s took %(c_White)s%(time)s%(c_Gold)s seconds to execute.%(c_reset)s",
		):
			if self.t.get(name) is None:
				msg = "You cannot stop the timer if it was not started"
				raise RuntimeError(msg)
			diff = time.perf_counter() - self.t[name]
			self._print_time(printhook, pattern, name, diff, color)

		def __call__(
			self,
			func=None,
			*,
			printhook=None,
			color=True,
			pattern="%(c_White)s%(name)s%(c_Gold)s took %(c_White)s%(time)s%(c_Gold)s seconds to execute.%(c_reset)s",
		):
			if func is None:
				return partial(timeit, printhook=printhook, pattern=pattern, color=color)
			if printhook is None:
				printhook = print

			@wraps(func)
			def wrapper(*args, **kwargs):
				start_time = time.perf_counter()
				result = func(*args, **kwargs)
				end_time = time.perf_counter()
				elapsed_time = end_time - start_time
				self._print_time(printhook, pattern, func.__name__, elapsed_time, color)
				return result

			return wrapper

	timeit = Timeit()


if True:  # \/ # @autorun, @instance

	@overload
	def autorun(func: F) -> F: ...

	@overload
	def autorun(*args, **kwargs) -> Callable[[F], F]: ...

	def autorun(*args, **kwargs):
		if len(args) == 1 and callable(args[0]):
			func = args[0]
			func()
			return func
		else:

			def decorator(func: F, args=args, kwargs=kwargs) -> F:
				func(*args, **kwargs)
				return func

			return decorator

	def instance(func):
		return func()


if True:  # \/ # @convert.stringify

	class Convert:
		def stringify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return str(func(*args, **kwargs))

			return wrapper

		def intify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return int(func(*args, **kwargs))

			return wrapper

		def floatify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return float(func(*args, **kwargs))

			return wrapper

		def boolify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return bool(func(*args, **kwargs))

			return wrapper

		def listify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return list(func(*args, **kwargs))

			return wrapper

		def setify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return set(func(*args, **kwargs))

			return wrapper

		def dictify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return dict(func(*args, **kwargs))

			return wrapper

		def tuplify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return tuple(func(*args, **kwargs))

			return wrapper

		def complexify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return complex(func(*args, **kwargs))

			return wrapper

		def reprify(self, func):
			@wraps(func)
			def wrapper(*args, **kwargs):
				return repr(func(*args, **kwargs))

			return wrapper

		def anyify(self, converter: Callable):
			def decorator(func):
				@wraps(func)
				def wrapper(*args, **kwargs):
					return converter(func(*args, **kwargs))

				return wrapper

			return decorator

	convert = Convert()

	def aconvert(converter: Callable, await_converter: bool = False):
		def decorator(afunc):
			@wraps(afunc)
			async def wrapper(*args, **kwargs):
				if await_converter:
					return await converter(await afunc(*args, **kwargs))
				else:
					return converter(await afunc(*args, **kwargs))

			return wrapper

		return decorator


if True:  # \/ # @printer

	def printer(*, printhook: Callable[[str], None] = print, passthrough: bool = True):
		def decorator(func: Callable):
			@wraps(func)
			def wrapper(*args, **kwargs):
				result = func(*args, **kwargs)
				printhook(result)
				if passthrough:
					return result

			return wrapper

		return decorator


if True:  # \/ # @copy_kwargs

	def copy_kwargs(func):
		"""Pack all of that func's kwargs into __kwargs and pass the kwarguments normally along with __kwargs kwarg which contains all the kwargs in a dict.

		This will error if func receives a literal "__kwargs" kwarg.
		```py
		>>> a(__kwargs=kwargs, **{"__kwargs": "some_value"})
		```
		```txt
		TypeError: __main__.a() got multiple values for keyword argument '__kwargs'
		```
		"""

		@wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, __kwargs=kwargs, **kwargs)

		return wrapper

	def copy_kwargs_sunder(func):
		"""Pack all of that func's kwargs into __kwargs and pass the kwarguments normally along with _kwargs kwarg which contains all the kwargs in a dict.

		This will error if func receives a literal "__kwargs" kwarg.
		```py
		>>> a(_kwargs=kwargs, **{"_kwargs": "some_value"})
		```
		```txt
		TypeError: __main__.a() got multiple values for keyword argument '_kwargs'
		```
		"""

		@wraps(func)
		def wrapper(*args, **kwargs):
			return func(*args, _kwargs=kwargs, **kwargs)

		return wrapper


if True:  # \/ # @skip_first_call

	def skip_first_call(func):
		func._first_call = True

		def wrapper(*args, **kwargs):
			if func._first_call:
				func._first_call = False
				return None
			else:
				return func(*args, **kwargs)

		return wrapper

	def skip_first_call_async(func):
		func._first_call = True

		async def wrapper(*args, **kwargs):
			if func._first_call:
				func._first_call = False
				return None
			else:
				return await func(*args, **kwargs)

		return wrapper


if True:  # \/ # @with_overrides

	def with_overrides(*overrides: str):
		"""For each of the passed in keyword argument names: if one was passed in by the user, use that one, else use `self.{name}`."""

		def decorator(func: Callable):
			def wrapper(self, *args, **kwargs):
				kws = kwargs.copy()
				for o in overrides:
					if o not in kwargs:
						kws[o] = getattr(self, o)

				return func(self, *args, **kws)

			return wrapper

		return decorator



class DynamicResponseError(Exception):
	"""Base class for all dynamic responses errors."""


class DynamicResponsePlaceholderMissingContextError(DynamicResponseError, TypeError):
	"""A placeholder requires a dependency which is missing."""


class DynamicResponseMissingPlaceholderError(DynamicResponseError, KeyError):
	"""Placeholder that was requested doesn't exist.

	(Is not raised if the error_on_missing_placeholder is set to False, in that case the literal text for the placeholder will be returned as the missing placeholder's return value.)

	```py
	execute = DynamicResponseBuilder(placeholders={
	  "test": lambda *args, **kwargs: "TEST",
	  "test2": lambda *args, **kwargs: "TEST2",
	})

	execute("{unknown_placeholder}") # Will raise ExecuteMissingPlaceholderError because 'unknown_placeholder' placeholder doesn't exist in the provided set.
	"""


class DynamicResponseSyntaxError(DynamicResponseError, ValueError):
	"""Syntax error in dynamic response's input text (for example mismatched parenthesis)."""


class DynamicResponseRecursionError(DynamicResponseError, RecursionError):
	"""Python's recursion limit was reached when evaluating a dynamic response. (for example placeholder tried to replace itself with itself so it got called again and again and you get the idea...)."""


class DynamicResponsePlaceholderInvalidReturnError(DynamicResponseError, ValueError):
	"""Placeholder returned an invalid value and raise_invalid_placeholder_return is set to True."""


class DynamicResponsePythonErrorInPlaceholerError(DynamicResponseError):
	"""Is raised when a Python error occurs within placeholder callable."""


class DynamicResponsePlaceholderTooFewArgumentsError(DynamicResponseError, TypeError):
	"""Placeholder requires more arguments than were provided.

	Example: {char} # char of.. what? you did not supply the argument.
	Simillar to python's 'TypeError: function() missing X required positional argument(s): ...'
	"""


class DynamicResponsePlaceholderTooManyArgumentsError(DynamicResponseError, TypeError):
	"""Placeholder requires strictly fewer arguments than were provided.

	This will probably be rarely used since in most cases the better approach is to ignore the extra arguments.

	Example: {char|0x4|0x3|0x2|0x1} # Too many arguments (assuming {char} doesn't support *args).
	Simillar to python's 'TypeError: function() takes X positional argument(s) but Y were given'
	"""
